//-------------------------------------------------------------------
//
//  gnuplot wrapper
//  
//-------------------------------------------------------------------
function gnu_plot()
{
    this.grid = false;                  // draw grid flag
    this.dt = 1;                        // dot type
    this.lw = 1;                        // line with
    this.pt = 7;                        // point type
    this.ps = 1;                        // point style
    this.color = "black";               // line color
    this.wait = 1500;                   // closing delay
    this.hold_on = false;               // plot or replot flag

    this.width = 800;
    this.height = 600;
    this.file_counter = 0;              // tmp file counter (for cleaning)
    this.file_list = [];                // tmp file list
    this.auto_refresh = false;          // refresh flag
    this.type = "line";                 // plot type
    this.debug = false;                 // show command
    this.window_counter = 0;            // window counter
    this.IsOpen = false;                // process flag
    this.terminal = "qt";              // terminal type
    this.tmp_ext = ".gpt";              // extension for tmp file
    // clear all previous temp files
    var tmp_list = Mi.GetDirList("*"+this.tmp_ext, Mi.GetTmpDirectory(), false);
    if (tmp_list)
    {
        var i;
        var len = tmp_list.length;
        for(i = 0; i < len; i++)
        {
            var filename = tmp_list[i];
            if (Mi.FileExists(filename))
                Mi.RemoveFile(filename);
        }
    }
    // start gnuplot
    this.GP = new Process("gnuplot.exe", "-persist");
}

gnu_plot.prototype.reset = function()
{
    this.grid = false;                  // draw grid flag
    this.dt = 1;                        // dot type
    this.lw = 1;                        // line with
    this.pt = 7;                        // point type
    this.ps = 1;                        // point style
    this.color = "black";               // line color
    this.hold_on = false;               // plot or replot flag
    this.type = "line";                 // plot type
}

gnu_plot.prototype.HoldOn = function(hold_on)
// set plot or replot mode
{
    this.hold_on = hold_on;
}

gnu_plot.prototype.Open = function()
// open gnuplot process
{
    if (this.IsOpen) return true;
    if (this.GP)
        this.IsOpen = this.GP.open();
    return this.IsOpen;
}

gnu_plot.prototype.Close = function()
// close process
{
    if (this.IsOpen)
    {
        Mi.MilliSleep(this.wait);       // and wait
        this.GP.send("quit");           // send quit command to gnuplot
        this.IsOpen = false;            //
        //this.GP.close();
        return true;
    }
    return false;
}

gnu_plot.prototype.Send = function(cmd)
// send command to gnuplot
{
    if (this.IsOpen)
    {
        if (this.debug) 
            print(cmd);
        this.GP.send(cmd);
        return true;
    }
    return false;
}

gnu_plot.prototype.SelectWindow = function(num, title)
// select a terminal window
{
    if (this.IsOpen && num >=0 && num < this.window_counter)
    {
        var cmd = "set term "+ this.terminal + " "+ num;
        if (title)
        {
            cmd = cmd + " title '" + title + "'";
        }
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
        return true;
    }
    return false;
}

gnu_plot.prototype.NewWindow = function(title)
// create a new window using current terminal type with title 'title'
{
    if (this.IsOpen)
    {
        var cmd = "set term "+ this.terminal + " "+ this.window_counter + " size "+this.width + ","+this.height;
        if (title)
        {
            cmd = cmd + " title '" + title + "'";
        }
        this.Send(cmd);
        // reset all styles
        this.Send("reset");
        this.Send("set autoscale");
        this.Send("set xlabel '' ");
        this.Send("set ylabel '' ");
        this.Send("set title '' ");
        this.Send("unset label");
        this.Send("unset key");
        this.Send("unset monochrome");
        this.window_counter++;
        return true;
    }
    return false;
}

gnu_plot.prototype.win_title = function(title)
{
    if (this.IsOpen && title != undefined)
    {
        var cmd = "set term "+ this.terminal + "  title '"+ title + "'";
        this.Send(cmd);
    }
}

gnu_plot.prototype.NewFigure = function(title)
// open a new empty window
{
    if (this.IsOpen)
    {
        this.NewWindow(title);
        this.Send("set xr[-1:1]");
        this.Send("set yr[-1:1]");
        this.Send("plot(0) notitle");
        this.Send("clear");
    }
}

gnu_plot.prototype.CheckPath = function(path_name)
// remove all '\' from path name
{
    var new_path = path_name;
    if (typeof(path_name) == "string")
    {
        new_path = "";
        var i;
        var n = path_name.length;
        for(i = 0; i < n; i++)
        {
            if (path_name[i] == "\\")
                new_path = new_path + '/';
            else
                new_path = new_path + path_name[i];
        }
    }
    return new_path;
}

gnu_plot.prototype.make_data = function(data)
// create a 2D matrix and save it to a text file
{
    if (this.IsOpen && data)
    {
        var path_name = Mi.GetTmpDirectory() + "/gp_tmp_data_" + this.file_counter + this.tmp_ext;
        path_name = this.CheckPath(path_name);
        var gp_file = new File(path_name, "w");
        if (gp_file.isOpen())
        {            
            var i, j;
            var sv = "";
            var nb_lig = data[0].size();
            var nb_col = data.length;
            for(i = 0; i < nb_lig; i++)
            {
                sv = "";
                for(j = 0; j < nb_col; j++)
                {
                    var v = data[j];
                    var value = v.get(i);
                    if (typeof(value) == "number")
                    {
                        sv = sv + value.toString();
                        if (j < nb_col -1) sv = sv + " ";
                    }
                    //else print("*** Make Data Error "+i+","+j);
                }             
                gp_file.writeLine(sv);                    
            }
            gp_file.close();
            if (this.debug)
                print("data file saved : "+path_name);
            this.file_list[this.file_counter] = path_name;
            this.file_counter++;
            return path_name;
        }
    }
    return undefined;
}

gnu_plot.prototype.refresh = function()
{
    if (this.IsOpen)
        this.Send("refresh");
}

gnu_plot.prototype.plot_2d = function(x_data, y_data, title, xlabel, ylabel)
// make a 2D plot
{
    if (this.IsOpen && x_data && y_data)
    {
        var M = new Array(2);
        M[0] = x_data;
        M[1] = y_data;
        
        //print("make_data : "+x_data.get(0)+" , "+y_data.get(0));
        
        var data_name = this.make_data(M);
        if (data_name)
        {
            var cmd = "plot ";
            if (this.hold_on) cmd = "replot ";
        
            var cmd = cmd + "\"" + data_name + "\"" + " with "+ this.type;
            if (title == undefined)  cmd = cmd + " notitle";
            else cmd = cmd + " title " + "\"" + title + "\"";
            
            cmd = this.plot_attr(cmd);            
            this.Send(cmd);

            if (title)
            {
                cmd = "set title '"+title+"'";
                //cmd = cmd + " font 'Helvetica'";
                this.Send(cmd);
            }
            if (xlabel)
            {
                cmd = "set xlabel '"+xlabel+"'";
                this.Send(cmd);
            }
            if (ylabel)
            {
                cmd = "set ylabel '"+ylabel+"'";
                this.Send(cmd);
            }        
            if (this.auto_refresh)
                this.Send("refresh");
            return true;
        }
    }
    return false;
}

gnu_plot.prototype.title = function(title, font, size)
{
    if (this.IsOpen && title)
    {
        var cmd = "set title '"+title+"'";
        if (font != undefined)
        {
            cmd = cmd + " font +\""+font;
            if (size != undefined)
                cmd = cmd +","+size;
            cmd = cmd +"\"";
        }
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
    }
}

gnu_plot.prototype.xlabel = function(label, font, size)
{
    if (this.IsOpen && label)
    {
        var cmd = "set xlabel '"+label+"'";
        if (font != undefined)
        {
            cmd = cmd + " font +\""+font;
            if (size != undefined)
                cmd = cmd +","+size;
            cmd = cmd +"\"";
        }
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
    }
}

gnu_plot.prototype.ylabel = function(label, font, size)
{
    if (this.IsOpen && label)
    {
        var cmd = "set ylabel '"+label+"'";
        if (font != undefined)
        {
            cmd = cmd + " font +\""+font;
            if (size != undefined)
                cmd = cmd +","+size;
            cmd = cmd +"\"";
        }
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
    }
}

gnu_plot.prototype.draw_x_axis = function()
{
    if (this.IsOpen)
    {
        this.Send("set xzeroaxis lt -1");
    }
}

gnu_plot.prototype.draw_y_axis = function()
{
    if (this.IsOpen)
    {
        this.Send("set yzeroaxis lt -1");
    }
}

gnu_plot.prototype.axis_2D = function(xmin, xmax, ymin ,ymax, title, xlabel, ylabel)
{
    if (this.IsOpen)
    {
        var cmd;
        if (xmin && xmax)
        {
            cmd = "set xr[" + xmin + ":" + xmax + "]";
            this.Send(cmd);
        }
        if (ymin && ymax)
        {
            cmd = "set yr[" + ymin + ":" + ymax + "]";
            this.Send(cmd);
        }
        this.title(title);
        this.xlabel(xlabel);
        this.ylabel(ylabel);
        if (this.auto_refresh)
            this.Send("refresh");
        return true;
    }
}

gnu_plot.prototype.plot_1d = function(rate, y_data, title, xlabel, ylabel)
{
    if (this.IsOpen && y_data)
    {
        var n = y_data.size();
        var time = new Float64Vector(n);
        var i;
        for(i = 0; i < n; i++)
        {
            var t = i/rate;
            time.set(i, t);
        }
        return this.plot_2d(time, y_data, title, xlabel, ylabel);
    }
    return false;
}

gnu_plot.prototype.save_plot = function(filename, ext)
{
    if (this.IsOpen && this.window_counter > 0 )
    {
        if (this.auto_refresh)
            this.Send("refresh");
            
        //var path = Mi.GetTmpDirectory() + "\\ms.gnu';
        var path_name = Mi.GetTmpDirectory() + "/gp_tmp.gnu";
        this.Send("save '"+ path_name +"'");

        var	gpSaveExt =  ["emf", "png",      "jpg",  "svg", "gif", "pdf",      "ps"];
        var	gpSaveType = ["emf", "pngcairo", "jpeg", "svg", "gif", "pdfcairo", "post"];        
        var Ext = "emf";        // default format
        var Type = Ext;         // and extension
        var size = gpSaveExt.length;
        if (typeof(ext) == "string")
        {
            var i;
            var t = -1;
            for(i = 0; i < size; i++)
            {
                if (ext == gpSaveExt[i]) t = i;
            }
            if (t >=0) 
            {
                Ext = ext;
                Type = gpSaveType[t];
            }
            else
                print("saveplot : using default format " + Ext);
        }
        var cmd = "set term "+ Type;
        this.Send(cmd);

        var Path = this.CheckPath(filename) + "."+Ext;
        if (this.debug) print("Saving plot : "+Path);

        var path = "\"" + Path + "\"";
        this.Send("set output " + path);
        this.Send("refresh");

        this.Send("load '"+ path_name +"'");
        return true;
    }
    return false;
}

gnu_plot.prototype.set_grid  = function(grid_on)
{
    this.grid = grid_on;
    if (this.IsOpen)
    {
        var cmd = "";
        if (grid_on) cmd = "set grid";
        else cmd = "unset grid";
        this.Send(cmd);
    }
}

gnu_plot.prototype.plot_attr  = function(cmd)
{
    cmd = cmd + " dt "+this.dt;
    cmd = cmd + " lc "+this.color;
    return cmd;
}

gnu_plot.prototype.plot_line = function(x1, y1, x2, y2)
{
    if (this.IsOpen)
    {
        var cmd = "set arrow from "+x1+","+y1+" to "+x2+","+y2+" nohead";
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
    }
}

gnu_plot.prototype.plot_line_value = function(value, title)
{
    if (this.IsOpen)
    {
        var cmd = "plot(";
        if (this.hold_on) cmd = "replot(";

        cmd = cmd + value + ") ";
        if (title) cmd = cmd + " title " + "\""+ title + "\"";
        else cmd = cmd + " notitle";

        cmd = this.plot_attr(cmd);
    
        this.Send(cmd);
        if (this.auto_refresh)
            this.Send("refresh");
        return true;
    }
    return false;
}

gnu_plot.prototype.plot_point = function(x, y)
{
    if (this.IsOpen)
    {
        var path_name = Mi.GetTmpDirectory() + "/gp_tmp_pt_" + this.file_counter +  this.tmp_ext;
        path_name = this.CheckPath(path_name);
        var gp_file = new File(path_name, "w");
        if (gp_file.isOpen())
        {            
            var cmd = "set pointsize "+ this.ps;
            this.Send(cmd);
            var sv = x.toString() + " " + y.toString();
            gp_file.writeLine(sv);
            gp_file.close();
            if (this.debug)
                print("data point file saved : "+path_name);
            this.file_list[this.file_counter] = path_name;
            this.file_counter++;

            cmd = "replot " + "\"" + path_name + "\"" + " with point pt "+ this.pt;
            this.Send(cmd);
            if (this.auto_refresh)
                this.Send("refresh");
        }
    }
}

gnu_plot.prototype.plot_points = function(x, y)
{
    if (this.IsOpen)
    {
        var path_name = Mi.GetTmpDirectory() + "/gp_tmp_pts_" + this.file_counter +  this.tmp_ext;
        path_name = this.CheckPath(path_name);
        var gp_file = new File(path_name, "w");
        if (gp_file.isOpen())
        {            
            var i;
            var len = 0;
            var sv;
            if (x.size == undefined)
            {
                len = x.length;
                for(i = 0; i < len; i++)
                {
                    sv = x[i].toString() + " " + y[i].toString();
                    gp_file.writeLine(sv);
                }
            }
            else
            {
                len = x.size();
                for(i = 0; i < len; i++)
                {
                    sv = x.get(i).toString() + " " + y.get(i).toString();
                    gp_file.writeLine(sv);
                }
            }
            gp_file.close();
            if (this.debug)
                print("data points file saved : "+path_name);
            this.file_list[this.file_counter] = path_name;
            this.file_counter++;
            var cmd = "replot " + "\"" + path_name + "\"" + " with point pt "+ this.pt;
            this.Send(cmd);
            if (this.auto_refresh)
                this.Send("refresh");
        }
    }
}
