//  mmatrix.c
//  Quelques fonctions utiles sur les tableaux de donnes (matrice 2D)
//
//  Gilles Dietrich, 2014

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "mmatrix.h"

void mtInitVector(struct MVECTOR* vec)
{
    if (vec != NULL)
    {
        vec->data = NULL;
        vec->row_count = 0;
    }
}

void mtCreateVector(struct MVECTOR* vec, int rows)
{
    if (vec != NULL && rows > 0)
    {
        vec->data = calloc(rows, sizeof(double));
        vec->row_count = rows;
    }
}

void mtClearVector(struct MVECTOR* vec)
{
    if (vec != NULL)
    {
        if (vec->data != NULL)
        {
            free(vec->data);
            vec->data = NULL;
            vec->row_count = 0;
        }
        mtInitVector(vec);
    }
}

#define MT_LINE_LEN    1024
#define MT_CELL_LEN     64

char    mtCEol      = '\n';
char    mtSEol[2]   = "\n";

char    mtCSep      = '\t';
char    mtSSep[2]   = "\t";


void mtSetSeparator(char c)
// assigne un nouveau dlimiteur de donnes (tabulation par dfaut)
{
    mtCSep = c;
    mtSSep[0] = c;
}

int mtReadTabDataFile(char* filename, double** mat, int* cols, int* rows)
// lecture d'un fichier texte contenant exclusivement des donnes numriques
// 'filename' est le nom du fichier
// 'mat' est un pointeur sur le tableau de donnes qui sera rempli
// si 'mat' est NULL, la fonction renvoie dans *cols et *rows le nombre de colonnes
// et de lignes du fichier.
// si 'mat' est non NULL, *cols et *rows doivent contenir les dimensions du tableau 'mat' et
// celui-ci contiendra les valeurs lues par cette fonction
// renvoie TRUE en cas de succs, sinon FALSE
{
    int ret = FALSE;
    FILE* f = fopen(filename, "r");                             // ouverture du fichier en lecture
    if (f!= NULL)                                               // test si tout est OK
    {
        char line[MT_LINE_LEN];                                 // une ligne du fichier texte
        int line_count = 0;
        int col_count = 0;                                      // compteur de colonnes
        while(TRUE)                                             // boucle sans fin
        {
            if (feof(f)) break;                                 // fin du fichier, quitter la bouble
            if ( fgets (line , MT_LINE_LEN , f) != NULL )       // obtient une ligne de texte
            {
                char* pEnd;
                char *p = strchr(line, mtCEol);                 // test si elle contient une fin de ligne
                if (p != NULL) *p = 0;                          // si oui on l'enlve
                p = strtok(line, mtSSep);                       // sparation des valeurs
                int local_col_count = 0;                        // compteur de colonne de cette ligne
                while (p != NULL)                               // sparer les valeurs sil elles existent
                {
                    if (mat != NULL && local_col_count < *cols && line_count < *rows)
                    {
                        double value = strtod(p, &pEnd);            // conversion en nombre rel 
                        mat[local_col_count][line_count] = value;   // remplissage du tableau (si il existe)
                    }
                    p = strtok(NULL, mtSSep);                   // valeur suivante
                    local_col_count++;                          // incrment du compteur de colonnes
                }
                if (mat == NULL)                                // on compte (on ne remplie pas le tableau)
                {
                    if (local_col_count > col_count) col_count = local_col_count;
                    *cols = col_count;                          // sauve les valeurs de colonnes
                    *rows = line_count+1;                       // et de ligne
                }
                line_count++;                                   // ligne suivante
            }
            else break;                                         // erreur : on sort de la boucle
        }
        fclose(f);                                              // fermeture du fichier
        if (mat == NULL)                                        // remplissage
        {
            if (line_count > 0 && col_count > 0) ret = TRUE;    // tout est OK
        }
        else ret = TRUE;
    }
    else
    {
        printf("Erreur lors de l'ouverture de %s : %s\n", filename, strerror(errno));
    }
    return ret;
}

void mtInitMatrix(struct MMATRIX* mat)
{
    if (mat != NULL)
    {
        mat->data = NULL;
        mat->row_count = 0;
        mat->col_count = 0;
    }
}

void mtCreateEmptyMatrix(struct MMATRIX* mat, int cols)
// allocation des colonnes de la matrice uniquement
{
    if (mat != NULL )
    {
        double** data = NULL;
        if (cols > 0)
        {
            data = (double**) malloc(cols*sizeof(double*));
        }
        else cols = 0;
        mat->col_count = cols;
        mat->row_count = 0;
        mat->data = data;
    }
}

void mtCreateMatrix(struct MMATRIX* mat, int cols, int rows)
// allocation d'un tableau de donnes (double) ayant 'col' colonnes et 
// 'rows' lignes.
{
    if (mat != NULL)
    {
        double** data = NULL;
        if (cols > 0 && rows > 0)
        {
            int i;
            data = (double**) malloc(cols*sizeof(double*));
            for(i = 0; i < cols; i++)
                data[i] = (double*) calloc(rows, sizeof(double));
        }
        else
        {
            cols =0;
            rows = 0;
        }
        mat->col_count = cols;
        mat->row_count = rows;
        mat->data = data;
    }
}

void mtClearMatrix(struct MMATRIX* mat)
// libration d'un tableau de donnes
{
    if (mat != NULL)
    if (mat->data != NULL)
    {
        for(int i = 0; i < mat->col_count; i++)
            free(mat->data[i]);
        free(mat->data);
        mat->data = NULL;
        mat->col_count = 0;
        mat->row_count = 0;
    }
}

int mtReadTabMatrix(char* filename, struct MMATRIX* mat)
// cette fonction ralise la lecture d'un fichier texte (sparateur = Tabulation
// 'filename' contient le nom du fichier  lire
{
    int ret = FALSE;
    if (mat != NULL)
    {
        int cols = 0;
        int rows = 0;
        if (mtReadTabDataFile(filename, NULL, &cols, &rows))              // obtention du nombre de lignes et de colonnes
        {
            mtCreateMatrix(mat, cols, rows);                                // cration du tableau
            mat->col_count = cols;
            mat->row_count = rows;
            if (!mtReadTabDataFile(filename, mat->data, &cols, &rows))      // lecture des valeurs
            {
                printf("Erreur de lecture de donnes\n");                   // erreur
                mtDeleteMatrix(mat);                                        // destruction du tableau
                mat = NULL;
            }
            else ret = TRUE;
        }
    }
    return ret;
}

int mtWriteTabMatrix(char* filename, struct MMATRIX* mat)
// cette fonction crit un tableau de donnes (double) de dimension cols, rows
// dans un fichier texte (tabulation)
{
    int ret = FALSE;
    if (filename != NULL && mat != NULL)
    if (mat->data != NULL && mat->col_count > 0 && mat->row_count > 0)
    {
        FILE* f = fopen(filename, "w");
        if (f != NULL)
        {
            char line[MT_LINE_LEN];
            char num_str[MT_CELL_LEN];
            int col, row;
            for(row = 0; row < mat->row_count; row++)
            {
                line[0] = 0;
                for(col = 0; col < mat->col_count; col++)
                {
                    double value = mat->data[col][row];
                    snprintf(num_str, MT_CELL_LEN-4, "%g", value);
                    if (col < mat->col_count-1) strcat(num_str, mtSSep);
                    strcat(line, num_str);
                }
                strcat(line, mtSEol);
                fputs(line, f);
            }
            fclose(f);
            ret = TRUE;
        }
        else 
            printf("Impossible de crer le fichier %s : %s \n", filename, strerror(errno));
    }
    return ret;
}
//------------------------------------------------------------------------------------
//  Matrices
//------------------------------------------------------------------------------------
int mtMatrixOk(struct MMATRIX* mat)
{
    int ret = FALSE;
    if (mat != NULL)
        if (mat->data != NULL && mat->row_count > 0 && mat->col_count > 0) ret = TRUE;
    return ret;
}

void mtPrintMatrix(struct MMATRIX* mat)
{
    if (mat != NULL)
    {
        if (mat->data != NULL)
        {
            for(int i = 0; i < mat->col_count; i++)
            {
                for(int j = 0; j < mat->row_count; j++)
                {
                    printf("data[%d][%d] = %g\t", i, j, mat->data[i][j]);
                }
                printf("\n");
            }
        }
        else printf("data = NULL\n");
    }
    else printf("NULL\n");
}

void mtZeroMatrix(struct MMATRIX* mat)
{
    if(mtMatrixOk(mat))
    {
        for(int i = 0; i < mat->col_count; i++)
            for(int j = 0; j < mat->row_count; j++)
                mat->data[i][j] = 0;
    }
}

int  mtIsSquare(struct MMATRIX* mat)
{
    if(mtMatrixOk(mat))
        if (mat->row_col == mat->col_count) return TRUE;
    return FALSE;
}

void mtSetId(struct MMATRIX* mat)
{
    if(mtMatrixOk(mat))
    {
        mtZeroMatrix(mat);
        int c = mat->row_count;
        if (mat->col_count < c) c = mat->col_count;
        for(int i = 0; i < c; i++)
            mat->data[i][i] = 1;
    }
}

int mtTransposeMatrix(struct MMATRIX* dest, struct MMATRIX* source)
{
    int ret = FALSE;
    if (dest != NULL && source != NULL)
    if (source->data != NULL && source->col_count > 0 && source->row_count > 0)
    {
        int n_cols = source->row_count;
        int n_rows = source->col_count;
        mtClearMatrix(dest);
        mtCreateMatrix(dest, n_cols, n_rows);
        for(int i = 0; i < n_cols; i++)
            for(int j = 0; j < n_rows; j++)
                dest->data[i][j] = source->data[j][i];
        ret = TRUE;
    }
    return ret;
}

int mtCopyMatrix(struct MMATRIX* dest, struct MMATRIX* source)
{
    int ret = FALSE;
    if (dest != NULL && mtMatrixOk(source))
    {
        int n_rows = source->row_count;
        int n_cols = source->col_count;
        mtClearMatrix(dest);
        mtCreateMatrix(dest, n_cols, n_rows);
        for(int i = 0; i < n_cols; i++)
            for(int j = 0; j < n_rows; j++)
                dest->data[i][j] = source->data[i][j];
        ret = TRUE;
    }
    return ret;
}

int mtScaleMatrix(struct MMATRIX* dest, struct MMATRIX* source, double scale)
{
    int ret = FALSE;
    if (mtCopyMatrix(dest, source))
    {
        int n_rows = dest->row_count;
        int n_cols = dest->col_count;
        for(int i = 0; i < n_cols; i++)
            for(int j = 0; j < n_rows; j++)
                dest->data[i][j] = dest->data[i][j]*scale;
        ret = TRUE;
    }
    return ret;
}

int mtMatrixSameSize(struct MMATRIX* s1,struct MMATRIX* s2)
{
    int ret = FALSE;
    if (mtMatrixOk(s1) && mtMatrixOk(s2))
    if (s1->row_count == s2->row_count && s1->col_count == s2->col_count)
    {
        ret = TRUE;
    }
    return ret;
}

int mtAddMatrix(struct MMATRIX* dest, struct MMATRIX* s1,struct MMATRIX* s2)
{
    int ret = FALSE;
    if (mtMatrixSameSize(s1, s2) && dest != NULL)
    {
        int n_rows = s1->row_count;
        int n_cols = s1->col_count;
        mtClearMatrix(dest);
        mtCreateMatrix(dest, n_cols, n_rows);
        for(int i = 0; i < n_cols; i++)
            for(int j = 0; j < n_rows; j++)
                dest->data[i][j] = s1->data[i][j] + s2->data[i][j];
        ret = TRUE;
    }
    return ret;
}

int mtSubMatrix(struct MMATRIX* dest, struct MMATRIX* s1,struct MMATRIX* s2)
{
    int ret = FALSE;
    if (mtMatrixSameSize(s1, s2) && dest != NULL)
    {
        int n_rows = s1->row_count;
        int n_cols = s1->col_count;
        mtClearMatrix(dest);
        mtCreateMatrix(dest, n_cols, n_rows);
        for(int i = 0; i < n_cols; i++)
            for(int j = 0; j < n_rows; j++)
                dest->data[i][j] = s1->data[i][j] - s2->data[i][j];
        ret = TRUE;
    }
    return ret;
}
//-----------------------------------------------------------------
//  Matrice carre
//-----------------------------------------------------------------
int mtCofactors(struct MMATRIX* fac, struct MMATRIX* mat)
{
    int ret = FALSE;
    if (fac != NULL)
    if (mtMatrixOk(mat))
    if (mat->row_count == mat->col_count && mat->row_count > 0)
    {        
        int order = mat->row_count;
        mtClearMatrix(fac);
        mtCreateMatrix(fac, order, order);
        
        int p, q, m, n, i, j;
        struct MMATRIX b;
        mtCreateMatrix(&b, order, order);

        for (q = 0; q < order; q++) 
        {
            for (p = 0; p < order; p++) 
            {
                m = 0;
                n = 0;
                for (i = 0; i < order; i++) 
                {
                    for (j = 0; j < order; j++) 
                    {
                        b.data[i][j] = 0;
                        if (i != q && j != p) 
                        {
                            b.data[m][n] = mat->data[i][j];
                            if (n < (order - 2))
                                n++;
                            else 
                            {
                                n = 0;
                                m++;
                            }
                        }
                    }
                }
                double det = mtDeterminant(&b, order - 1);
                fac->data[q][p] = pow(-1, q + p) * det;
            }
        }
        mtClearMatrix(&b);
        ret = TRUE;
    }
    return ret;
}

double mtDeterminant(struct MMATRIX* mat, int order)
{
    double det = 0;
    if (mtMatrixOk(mat))
    if (mat->row_count == mat->col_count && mat->row_count > 0)
    {
        if (order <=0) order = mat->row_count;
        if (order == 1) det = mat->data[0][0];
        else
        {
            struct MMATRIX tmp;
            mtCreateMatrix(&tmp, order, order);
            int i, j, m, n, c;
            double s = 1;

            for (c = 0; c < order; c++) 
            {
                m = 0;
                n = 0;
                for (i = 0; i < order; i++) 
                {
                    for (j = 0; j < order; j++) 
                    {
                        tmp.data[i][j] = 0;
                        if (i != 0 && j != c) 
                        {
                            tmp.data[m][n] = mat->data[i][j];
                            if (n < (order - 2))
                                n++;
                            else 
                            {
                                n = 0;
                                m++;
                            }
                        }
                    }
                }
                det = det + s * (mat->data[0][c] * mtDeterminant(&tmp, order - 1));
                s = -1 * s;
            }
            mtClearMatrix(&tmp);
        }
    }    
    return det;
}

int mtInverseSquareMatrix(struct MMATRIX* dest, struct MMATRIX* mat)
{
    int ret = FALSE;
    if (mtMatrixOk(mat) && dest != NULL)
    if (mat->row_count == mat->col_count && mat->row_count > 0)
    {
        int order = mat->row_count;
        double det = mtDeterminant(mat, order);
        if (det != 0)
        {
            struct MMATRIX fac, b;
            mtInitMatrix(&fac);
            mtInitMatrix(&b);
            if (mtCofactors(&fac, mat))
            {
                int i, j;
                mtCreateMatrix(&b, order, order);
                mtClearMatrix(dest);
                mtCreateMatrix(dest, order, order);
                for (i = 0; i < order; i++) 
                    for (j = 0; j < order; j++) 
                        b.data[i][j] = fac.data[j][i];
                for (i = 0; i < order; i++) 
                    for (j = 0; j < order; j++) 
                    dest->data[i][j] = b.data[i][j] / det;
            }
            mtClearMatrix(&fac);
            mtClearMatrix(&b);
            ret = TRUE;
        }
    }
    return ret;
}
//-----------------------------------------------------------------
//  Vecteur
//-----------------------------------------------------------------
int mtVectorOk(struct MVECTOR* v)
{
    int ret = FALSE;
    if (v != NULL)
    if (v->data != NULL && v->row_count > 0) ret = TRUE;
    return ret;
}

void mtZeroVector(struct MVECTOR* v)
{
    if (mtVectorOk(v))
    {
        for(int i = 0; i < v->row_count; i++)
            v->data[i] = 0;
    }
}

void mtPrintVector(struct MVECTOR* v)
{
    if (v != NULL)
    {
        if (v->data !=NULL)
        {
            for(int i = 0; i < v->row_count; i++)
            {
                printf("data[%d] = %g\n", i, v->data[i]);
            }
        }
        else printf("data = NULL\n");
    }
    else printf("NULL\n");
}

double mtDotProduct(struct MVECTOR* v1, struct MVECTOR* v2)
{
    double p = 0;
    if (mtVectorOk(v1) && mtVectorOk(v2))
    {
        for(int i = 0; i < v1->row_count; i++)
        {
            p+= v1->data[i]*v2->data[i];
        }
    }
    return p;
}

double mtNormVector(struct MVECTOR* v)
{
    double n = 0;
    if (mtVectorOk(v))
    {
        for(int i = 0; i < v->row_count; i++)
            n+= (v->data[i]*v->data[i]);
        n = sqrt(n);
    }
    return n;
}

int mtNormalizeVector(struct MVECTOR* v)
{
    int ret = FALSE;
    if (mtVectorOk(v))
    {
        double n = mtNormVector(v);
        if (n > MT_EPS)
        {
            for(int i = 0; i < v->row_count; i++)
                v->data[i]/=n;
            ret = TRUE;
        }
    }
    return ret;
}

int mtScaleVector(struct MVECTOR* v, double s)
{
    int ret = FALSE;
    if (mtVectorOk(v))
    {
        for(int i = 0; i < v->row_count; i++)
            v->data[i]*=s;
        ret = TRUE;
    }
    return ret;
}

int mtAddVector(struct MVECTOR* dest, struct MVECTOR* s1, struct MVECTOR* s2)
{
    int ret = FALSE;
    if (mtVectorOk(s1) && mtVectorOk(s2) && dest != NULL)
    if (s1->row_count == s2->row_count)
    {
        mtClearVector(dest);
        int rows = s1->row_count;
        mtCreateVector(dest, rows);
        for(int i = 0; i < rows; i++)
            dest->data[i] = s1->data[i]+s2->data[i];
        ret = TRUE;
    }
    return ret;
}

int mtSubVector(struct MVECTOR* dest, struct MVECTOR* s1, struct MVECTOR* s2)
{
    int ret = FALSE;
    if (mtVectorOk(s1) && mtVectorOk(s2) && dest != NULL)
    if (s1->row_count == s2->row_count)
    {
        mtClearVector(dest);
        int rows = s1->row_count;
        mtCreateVector(dest, rows);
        for(int i = 0; i < rows; i++)
            dest->data[i] = s1->data[i]+s2->data[i];
        ret = TRUE;
    }
    return ret;
}

int mtMid(struct MVECTOR* dest, struct MVECTOR* s1, struct MVECTOR* s2)
{
    int ret = FALSE;
    if (mtVectorOk(s1) && mtVectorOk(s2) && dest != NULL)
    if (s1->row_count == s2->row_count)
    {
        mtClearVector(dest);
        int rows = s1->row_count;
        mtCreateVector(dest, rows);
        for(int i = 0; i < rows; i++)
            dest->data[i] = (s1->data[i] + s2->data[i])/2.0;
        ret = TRUE;
    }
    return ret;
}

double mtDistance(struct MVECTOR* s1, struct MVECTOR* s2)
{
    double d = 0;
    if (mtVectorOk(s1) && mtVectorOk(s2))
    if (s1->row_count == s2->row_count)
    {
        struct MVECTOR tmp;
        mtInitVector(&tmp);
        if (mtSubVector(&tmp, s1, s2))
        {
            d = mtNormVector(&tmp);
            mtClearVector(&tmp);
        }
    }
    return d;
}

int mtCrossProduct3D(struct MVECTOR* dest, struct MVECTOR* s1, struct MVECTOR* s2)
{
    int ret = FALSE;
    if (mtVectorOk(s1) && mtVectorOk(s2) && dest != NULL)
    if (s1->row_count == 3)
    {
        mtClearVector(dest);
        int rows = s1->row_count;
        mtCreateVector(dest, rows);
        dest->data[0] = s1->data[1]*s2->data[2] - s1->data[2]*s2->data[1];
        dest->data[1] = s1->data[2]*s2->data[0] - s1->data[0]*s2->data[2];
        dest->data[2] = s1->data[0]*s2->data[1] - s1->data[1]*s2->data[0];
        ret = TRUE;
    }
    return ret;
}

int mtCopyVector(struct MVECTOR* dest, struct MVECTOR* source)
{
    int ret = FALSE;
    if (mtVectorOk(source) && dest != NULL)
    {
        mtClearVector(dest);
        int rows = source->row_count;
        mtCreateVector(dest, rows);
        for(int i = 0; i < rows; i++)
            dest->data[i] = source->data[i];
        ret = TRUE;
    }
    return ret;
}
//--------------------------------------------------------------------------------
//  Vecteurs et matrices
//--------------------------------------------------------------------------------
int mtMatVecMul(struct MVECTOR* dest, struct MVECTOR* source, struct MMATRIX* mat)
{
    int ret = FALSE;
    if (mtMatrixOk(mat) && mtVectorOk(source) && dest != NULL)
    if (source->row_count == mat->col_count)
    {
        mtClearVector(dest);
        int rows = source->row_count;
        mtCreateVector(dest, rows);
        for(int i=0; i < mat->row_count; i++)
        {
            double value = 0.0;
            for(int j = 0; j < mat->col_count; j++)
                value+=  source->data[j]*mat->data[j][i];
            dest->data[i] = value;
        }
        ret = TRUE;
    }
    return ret;
} 

int mtVectorsToMatrix3D(struct MMATRIX* dest, struct MVECTOR* v1, struct MVECTOR* v2, int a1, int a2)
//	Calcul d'une matrice de rotation  partir de deux vecteurs non colinaires
//	l'axe a1 correspond au prmier vecteur s1, a2 l'axe secondaire  s2
{
    int ret = FALSE;
    if (dest != NULL && mtVectorOk(v1) && mtVectorOk(v2))
    if (v2->row_count == 3 && v1->row_count == 3)
    if ((a1 != a2) && a1 >=0 && a1 <3 && a2 >=0 && a2 < 3)
    {
        mtClearMatrix(dest);
        struct MVECTOR a, c, b;
        mtInitVector(&a);
        mtInitVector(&b);
        mtInitVector(&c);
        
        if (!mtCopyVector(&a, v1)) goto end;
        if (!mtNormalizeVector(&a)) goto end;
        
        int a3 = 3-a1-a2;
        if ( (a1 == 0 && a2 == 1) || (a1 == 1 && a2 == 2) || (a1 == 2 && a2 == 0) )
        {
            if (!mtCrossProduct3D(&c, &a, v2))  goto end;
            if (!mtNormalizeVector(&c))         goto end;
            if (!mtCrossProduct3D(&b, &c, &a))  goto end;
        }
        else
        {
            if (!mtCrossProduct3D(&c, v2, &a))  goto end;
            if (!mtNormalizeVector(&c))         goto end;
            if (!mtCrossProduct3D(&b, &a, &c))  goto end;
        }
        mtCreateMatrix(dest, 3, 3);
        for(int i = 0; i < 3; i++)
        {
            dest->data[a1][i] = a.data[i];
            dest->data[a2][i] = b.data[i];
            dest->data[a3][i] = c.data[i];
        }
        ret = TRUE;
end:
        mtClearVector(&a);
        mtClearVector(&b);
        mtClearVector(&c);
        
    }
    return ret;
}

int mtFrameToMatrix3D(struct MMATRIX* dest, struct MVECTOR* P1, struct MVECTOR* P2, struct MVECTOR* P3, int a1, int a2)
{
    int ret = FALSE;
    struct MVECTOR v1, v2;
    mtInitVector(&v1);
    mtInitVector(&v2);
    if (mtSubVector(&v1, P2, P1))
    {
        if (mtSubVector(&v2, P3, P1))
        {
            ret = mtVectorsToMatrix3D(dest, &v1, &v2, a1, a2);
        }
    }
    mtClearVector(&v1);
    mtClearVector(&v2);
    return ret;
}

int mtToLocal3D(struct MVECTOR* dest, struct MVECTOR* source, struct MVECTOR* P0, struct MVECTOR* P1, struct MVECTOR* P2, struct MVECTOR* P3, int a1, int a2)
// Calcul des coordonnes d"un point 3D 'dest' en coordonnes Locales
// si on connait ses coordonnes 'source' dans le repre global constitu par :
// P1, P2 et P3 dfinissent les 3 points d'un repre
// axe1 et axe2 dtermine les axes calcule (0:X, 1:Y, 2:Z)
// Point correspond au coordonnes du centre du repre local
// Origine (0,0,0) correspond aux coordonnes du repre global
// en retour le point 'dest' (correspondant  'source') est calcul
{
    int ret = FALSE;
    if (mtVectorOk(source) && mtVectorOk(P0) && mtVectorOk(P1) && mtVectorOk(P2) && mtVectorOk(P3) && dest != NULL)
    {        
        struct MVECTOR  P;
        struct MMATRIX  R, I;
        mtInitVector(&P);
        mtInitMatrix(&R);
        mtInitMatrix(&I);
        if (mtFrameToMatrix3D(&R, P1, P2, P3, a1, a2))
        {
            if (mtInverseSquareMatrix(&I, &R))
            {
                mtCreateVector(&P, 3);
                for(int i = 0; i < 3; i++)
                    P.data[i] = source->data[i] - P0->data[i];
                ret = mtMatVecMul(dest, &P, &I);
            }
        }
        mtClearVector(&P);
        mtClearMatrix(&R);
        mtClearMatrix(&I);
    }
    return ret;
}

int mtToMain3D(struct MVECTOR* dest, struct MVECTOR* source, struct MVECTOR* P0, struct MVECTOR* P1, struct MVECTOR* P2, struct MVECTOR* P3, int a1, int a2)
// Calcul des coordonnes d"un point 3D 'dest' en coordonnes Globales
// si on connait ses coordonnes 'source' dans le repre local constitu par :
// P1, P2 et P3 dfinissent les 3 points d'un repre
// axe1 et axe2 dtermine les axes calcule (0:X, 1:Y, 2:Z)
// Point correspond au coordonnes du centre du repre local
// Origine (0,0,0) correspond aux coordonnes du repre global
// en retour le point 'dest' (correspondant  'source') est calcul
{
    int ret = FALSE;
    if (mtVectorOk(source) && mtVectorOk(P0) && mtVectorOk(P1) && mtVectorOk(P2) && mtVectorOk(P3) && dest != NULL)
    {
        struct MVECTOR  P;
        struct MMATRIX  R;
        mtInitVector(&P);
        mtInitMatrix(&R);
        if (mtFrameToMatrix3D(&R, P1, P2, P3, a1, a2))
        {
            if (mtMatVecMul(&P, source, &R))
            {
                mtClearVector(dest);
                mtCreateVector(dest, 3);
                for(int i = 0; i < 3; i++)
                    dest->data[i]=P.data[i] + P0->data[i];
                ret = TRUE;
            }
        }
        mtClearVector(&P);
        mtClearMatrix(&R);
    }
    return ret;
}

//------------------------------------------------------------------------------------
//  test de cette lib
//------------------------------------------------------------------------------------
/*
//#define TEST

#ifdef TEST

#define SIZE    3
void main()
{
    struct MVECTOR v1, v2, v3, v4;     // dclaration d'un vecteur
    mtInitVector(&v1);             // initialisation
    mtInitVector(&v2);             // initialisation
    mtInitVector(&v3);             // initialisation
    mtInitVector(&v4);             // initialisation

    mtCreateVector(&v1, SIZE);     // cration des donnes
    mtCreateVector(&v2, SIZE);     // cration des donnes
    mtCreateVector(&v4, SIZE);     // cration des donnes
    for(int i = 0; i < SIZE; i++)
    {
        v1.data[i] = i;
        v2.data[i] = i+SIZE;
    }
    double dist = mtDistance(&v2, &v1);
    printf("Distance = %g\n", dist);
    mtMid(&v3, &v1, &v2);
    
    mtNormalizeVector(&v1);
    mtNormalizeVector(&v2);
    mtCrossProduct3D(&v3, &v1, &v2);

    mtPrintVector(&v1);
    mtPrintVector(&v2);
    mtPrintVector(&v3);

    double dp = mtDotProduct(&v1, &v2);
    mtCopyVector(&v2, &v1);
    printf("DP = %g\n", dp);
    mtPrintVector(&v2);
    
    struct MMATRIX m1, m2, m3, m4;        // idem pour la matrice
    mtInitMatrix(&m1);
    mtInitMatrix(&m2);
    mtInitMatrix(&m3);
    mtInitMatrix(&m4);

    mtCreateMatrix(&m1, SIZE, SIZE);
    mtSetId(&m1);
    mtTransposeMatrix(&m2,&m1);
    mtPrintMatrix(&m1);
    printf("Transpose  : \n");
    mtPrintMatrix(&m2);

    v1.data[0] = 0;
    v1.data[1] = 1;
    v1.data[2] = 0;
    v2.data[0] = 0;
    v2.data[1] = 0;
    v2.data[2] = 1;
    if (mtVectorsToMatrix3D(&m3, &v1, &v2, 1, 2))
    {
        printf("Rotation\n");
        mtPrintMatrix(&m3);
        if (mtInverseSquareMatrix(&m4, &m3))
        {
            printf("Matrice Inverse\n");
            mtPrintMatrix(&m4);
        }
        double det = mtDeterminant(&m3, 3);
        printf("Determinant = %g\n", det);
        if (mtCofactors(&m4, &m3))
        {
            printf("Cofacteur\n");
            mtPrintMatrix(&m4);
        }
    }
    
    v1.data[0] = 0;
    v1.data[1] = 0;
    v1.data[2] = 0;
    
    v2.data[0] = 1;
    v2.data[1] = 1;
    v2.data[2] = 1;
    
    v3.data[0] = 0;
    v3.data[1] = 0.5;
    v3.data[2] = 0.5;
    
    v4.data[0] = sqrt(3.0);
    v4.data[1] = 0;
    v4.data[2] = 0;
    
    struct MVECTOR dst;
    mtInitVector(&dst);
    
    if (mtToMain3D(&dst, &v4, &v1, &v1, &v2, &v3, 0, 1))
    {
        printf("mtToMain\n");
        mtPrintVector(&dst);
    }

    v4.data[0] = 1;
    v4.data[1] = 1;
    v4.data[2] = 1;
    if (mtToLocal3D(&dst, &v4, &v1, &v1, &v2, &v3, 0, 1))
    {
        printf("mtToLocal\n");
        mtPrintVector(&dst);
    }
    mtClearVector(&dst);
    
    mtClearVector(&v1);
    mtClearVector(&v2);
    mtClearVector(&v3);
    mtClearVector(&v4);

    mtCopyMatrix(&m1, &m4);
    mtScaleMatrix(&m1, &m4, 0.5);
    mtAddMatrix(&m2, &m1, &m4);
    mtSubMatrix(&m3, &m1, &m4);

    mtPrintMatrix(&m2);
    mtPrintMatrix(&m3);

    mtClearMatrix(&m1);
    mtClearMatrix(&m2);
    mtClearMatrix(&m3);
    mtClearMatrix(&m4);
}
#endif
*/